@@ -1,6 +1,6 @@
 from __future__ import annotations
 
-__version__ = "4.0.0"
+__version__ = "4.4.3"
 
 
 import asyncio
@@ -19,11 +19,13 @@ from .bluez import (  # noqa: F401
     Allocations,
     BleakSlotManager,
     _get_properties,
+    _get_services_cache,
     clear_cache,
     device_source,
     get_connected_devices,
     get_device,
     get_device_by_adapter,
+    path_from_ble_device,
     wait_for_device_to_reappear,
     wait_for_disconnect,
 )
@@ -121,7 +123,18 @@ TRANSIENT_ERRORS_MEDIUM_BACKOFF = {
 
 DEVICE_MISSING_ERRORS = {"org.freedesktop.DBus.Error.UnknownObject"}
 
-OUT_OF_SLOTS_ERRORS = {"available connection", "connection slot"}
+# ESP_GATT_CONN_CONN_CANCEL (0x100) indicates the ESP32 rejected the connection due to
+# limited resources (HCI error 0x0d). This happens when ESPHome incorrectly marks a
+# connection slot as free at ESP_GATTC_DISCONNECT_EVT instead of waiting for
+# ESP_GATTC_CLOSE_EVT, causing a race where we try to use a slot that isn't actually
+# available yet. The 4-second backoff gives time for the slot to truly become available
+# and for the state to sync with Home Assistant.
+# See: https://github.com/espressif/esp-idf/issues/17452
+OUT_OF_SLOTS_ERRORS = {
+    "available connection",
+    "connection slot",
+    "ESP_GATT_CONN_CONN_CANCEL",
+}
 
 TRANSIENT_ERRORS = {
     "le-connection-abort-by-local",
@@ -129,7 +142,6 @@ TRANSIENT_ERRORS = {
     "ESP_GATT_CONN_FAIL_ESTABLISH",
     "ESP_GATT_CONN_TERMINATE_PEER_USER",
     "ESP_GATT_CONN_TERMINATE_LOCAL_HOST",
-    "ESP_GATT_CONN_CONN_CANCEL",
 } | OUT_OF_SLOTS_ERRORS
 
 # Currently the same as transient error
@@ -223,6 +235,92 @@ def ble_device_description(device: BLEDevice) -> str:
     return base_name
 
 
+async def _has_valid_services_in_cache(device: BLEDevice) -> bool:
+    """Check if the device has valid services in cache.
+
+    For non-Linux platforms (macOS, ESPHome proxies), the cache is always valid
+    and we return True to allow using cached services.
+
+    For Linux/BlueZ, this function validates that cached services are still valid
+    by checking if they exist in the BlueZ manager's properties. Due to race
+    conditions, the properties and services cache can become out of sync. If
+    properties have disappeared but the services cache still contains them, the
+    cache is stale and should not be used.
+
+    Returns:
+    - True for non-Linux platforms (cache always valid)
+    - True for Linux if all cached services are still present in D-Bus properties
+    - False for Linux if cache is stale or unavailable
+    """
+    if not IS_LINUX:
+        # Cache is always valid on macOS and ESPHome proxies
+        return True
+
+    # Check if this is a BlueZ device
+    if not (device_path := path_from_ble_device(device)):
+        # Not a BlueZ device (might be ESPHome proxy), cache is valid
+        return True
+
+    # Get the services cache
+    if not (services_cache := await _get_services_cache()):
+        _LOGGER.debug(
+            "%s - %s: No services cache available, cannot validate",
+            device.name or "Unknown",
+            device.address,
+        )
+        return False
+
+    # Check if services are cached for this device
+    if not (cached_services := services_cache.get(device_path)):
+        _LOGGER.debug(
+            "%s - %s: No cached services found for device path %s",
+            device.name or "Unknown",
+            device.address,
+            device_path,
+        )
+        return False
+
+    # Get current properties to check if cached services are still present
+    if not (properties := await _get_properties()):
+        _LOGGER.debug(
+            "%s - %s: Could not get properties to validate cache",
+            device.name or "Unknown",
+            device.address,
+        )
+        return False
+
+    service_count = len(cached_services.services)
+    if not service_count:
+        _LOGGER.debug(
+            "%s - %s: No cached services found for device path %s",
+            device.name or "Unknown",
+            device.address,
+            device_path,
+        )
+        return False
+
+    # Check if all cached services are still present in properties
+    # The cached_services is a BleakGATTServiceCollection object
+    for service in cached_services:
+        if service.obj[0] not in properties:
+            # Service is in cache but not in properties (not on the bus)
+            _LOGGER.debug(
+                "%s - %s: Cached service %s not found in properties, cache invalid",
+                device.name or "Unknown",
+                device.address,
+                service,
+            )
+            return False
+
+    _LOGGER.debug(
+        "%s - %s: All %d cached services are valid and present in properties",
+        device.name or "Unknown",
+        device.address,
+        service_count,
+    )
+    return True
+
+
 def calculate_backoff_time(exc: Exception) -> float:
     """Calculate the backoff time based on the exception."""
 
@@ -349,7 +447,12 @@ async def establish_connection(
         # device.
         device = devices[0]
 
-    client = client_class(device, disconnected_callback=disconnected_callback, **kwargs)
+    client = client_class(
+        device,
+        disconnected_callback=disconnected_callback,
+        _is_retry_client=True,
+        **kwargs,
+    )
 
     while True:
         attempt += 1
@@ -363,10 +466,14 @@ async def establish_connection(
 
         try:
             async with asyncio_timeout(BLEAK_SAFETY_TIMEOUT):
+                # Only use cache if we have valid services in the cache
+                should_use_cache = use_services_cache or bool(cached_services)
+                if should_use_cache:
+                    should_use_cache = await _has_valid_services_in_cache(device)
+
                 await client.connect(
                     timeout=BLEAK_TIMEOUT,
-                    dangerous_use_bleak_cache=use_services_cache
-                    or bool(cached_services),
+                    dangerous_use_bleak_cache=should_use_cache,
                 )
                 if debug_enabled:
                     _LOGGER.debug(
@@ -533,12 +640,16 @@ async def restore_discoveries(scanner: BleakScanner, adapter: str) -> None:
         return
     backend = scanner._backend
     before = len(backend.seen_devices)
+    details: dict[str, Any]
     backend.seen_devices.update(
         {
-            address: (history.device, history.advertisement_data)
-            for address, history in load_history_from_managed_objects(
+            path: (device, history.advertisement_data)
+            for history in load_history_from_managed_objects(
                 properties, adapter
-            ).items()
+            ).values()
+            if (device := history.device)
+            and (details := device.details)
+            and (path := details.get("path"))
         }
     )
     _LOGGER.debug(
